/***************************************************************************
* atpg/compactor/compactor.cc     
* This file implements the compactor
*
* $Id: compactor.cc,v 1.2 2003/06/24 20:55:49 mdsuresh Exp $
*
* Copyright (C) 2003 Suresh kumar Devanathan <mdsuresh@sourceforge.net>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* Suggestions:          atpg-dev@sourceforge.net
* Bugs:                 atpg-bugs@sourceforge.net
*
* For more information, please see the Blitz++ Home Page:
*    http://sourceforge.net/projects/atpg/
*
***************************************************************************
* $Log: core.cc,v $
* Revision 1.0  2003/06/24 20:22:49  mdsuresh
* Updated copyright date in headers.
**/


#ifndef __FAULT_COMPACTOR__
#define __FAULT_COMPACTOR__ "1x"



namespace FaultAnalysis {
	namespace Simulator {

		using std::map;
		using std::greater_equal;
		using std::less;
		using std::multimap;
		using std::priority_queue;
		using std::deque;
		using std::queue;
		using std::ostream;
		using std::ofstream;
		using std::list;
		using std::vector;
		using Helper::FaultAnalysis::LinkAndLineIndex;
		using Helper::VectorCollection;
		using Helper::FaultAnalysis::MaxInputLines;


		struct Block {
			 enum Phase {
				 VALIDATION,
				 REFINEMENT,
				 PRUNE,
				 COMMIT
			 }  phase;
			 int min, max,opt, top;
			 int base, i;
			 bool pruning, refinement;

			 Block()
			 {
				 min  =0;
				 max = 0;
				 opt = 0;
				 base = 0;
				 top = 0;
				 i = 0;
				 pruning = false;
				 refinement = false;
			 }

			 Block( size_t u_detected)
			 {
				 top = max = base = min = u_detected;
				 i = 0;
				 phase = VALIDATION;
			 }

			 void Search(bool fault_detected, const vector<bool>& removed)
			 {



				 switch( phase)
				 {
				 case VALIDATION:

					 //std::cout << "V:";

					 if(!fault_detected)
						{

							do 
							{

								min = max;
								max = base - 1 << i;  i++;


								//std::cout << max  << " - " << min << " "; 
								if( max <  0)
								{
									max = 0;
									min = max;
									top =0;
									phase  = REFINEMENT;
									opt = 0;
									i =0;
								}
						} 
							while(NoRemovedLeft(removed, max,base) && phase==VALIDATION);
						}
						else
						{

							top = max;
							min = max;
							phase = REFINEMENT;
							i = 0;
							//if( max < (min - 1))
							//	opt = (max + min) / 2;
							//else
							//{
							//	phase = COMMIT;
							//}

						}


						break;

				 case REFINEMENT:

				 if(fault_detected)
						{

							do 
							{

								min = max;
								max = top + 1 << i;  i++;


								//std::cout << max  << " - " << min << " "; 
								if( max >  base)
								{
									max = base;
									phase  = COMMIT;
									opt = 0;
								}
							} 
							while(NoRemovedLeft(removed, top, max) && phase==REFINEMENT);
						}
						else
						{

							phase = COMMIT;
							max = min;
							//if( max < (min - 1))
							//	opt = (max + min) / 2;
							//else
							//{
							//	phase = COMMIT;
							//}

						}


						break;


						////do
					 //   
						//{

						//	if( max >= (min - 1))
						//	{
						//		phase = COMMIT;
						//		break;
						//	}

						//	if(!fault_detected)
						//	{
						//		min = opt;
						//	}
						//	else 
						//		max = opt;

						//	if( max < (min - 1))
						//		opt = (max + min)/2;
						//	else
						//	{
						//		phase = COMMIT;
						//		break;
						//	}

						//} 

						//
						///*
						//if( fault_detected && refinement)
						//{
						//	if(Skip( removed, opt))
						//		 opt = opt - 1;

						//	if( opt >= (min - 1))
						//	{
						//		max = opt;
						//		phase = PRUNE;
						//		break;
						//	}
						//}
						//*/

						//

						 
						//while(NoRemovedLeft(removed, opt, (opt+min)/2) && refinement && phase == REFINEMENT  && fault_detected);

						refinement = true;

						//std::cout << "R:" << opt << " - " << base << " :";

						break;

				 case PRUNE:

					 if(fault_detected)
						{
							pruning = true;

							if(!Skip( removed, max))
								max++;


						}
						else
						{
							phase = COMMIT;
							if(pruning) max = max - 1;
							pruning = false;
						}
						if( max > base)
						{
							max = base;
							phase = COMMIT;
							pruning = false;
						}

						//std::cout << "P:" << max << " - " << base << " :";
						break;

				 case COMMIT:
					 //std::cout << "C:" << max << " - " << base << " :";
					 break;

				 }




			 }

			 bool NoRemovedLeft( const vector<bool>& removed, int min, int max)
			 {
				 if( max < min )
					 return false;
				 for(int i = min; i <= max; i++)
					 if(removed[i])
						 return false;
				 return true;
			 }

			 bool Skip( const vector<bool>& removed, int& max)
			 {
				 bool skipped = false;
							while(max<=base)
							{
								if(removed[max])
									break;
								  max++;
								  skipped = true;
							}
					return skipped;

			 }

			 bool SkipReverse( const vector<bool>& removed, int& max)
			 {
				 bool skipped = false;
							while(max>=0)
							{
								if(removed[max])
									break;
								  max--;
								  skipped = true;
							}
					return skipped;

			 }


			 void PrepareSequence( vector<bool> & omitted)
			 {
				 
				 switch( phase)
				 {
				 case VALIDATION:
					 set_Sequence( omitted, max, base);
					 break;
				 case REFINEMENT:
					 set_Sequence( omitted, opt, base);
					 break;
				 case PRUNE:
				 case COMMIT:
					 set_Sequence( omitted, max, base);
					 break;
				 default:
					 assert(false);
				 }

				 
                 
			 }

			 bool IsCommitable() const
			 {
				 return phase == COMMIT;

			 }

			 void Commit( vector<bool>& omitted)
			 {
				 set_Sequence( omitted, max, base);
			 }

			 void set_Sequence( vector<bool>& omitted, size_t begin, size_t end)
			 {
                 for( size_t j = begin; j <= end; j++)
					  omitted[j] = false;
			 }
			 

		};


		template< class FaultCollection > 
		struct BlockedRestore
		{
			typedef _ag_imp_typename FaultCollection::Link  Link;
			typedef map< Link , Block >  LinkList;
			LinkList links;

			FaultCollection get_FaultList() const
			{
				FaultCollection return_list;

				for( _ag_imp_typename LinkList::const_iterator it = links.begin(); it != links.end(); it++)
					return_list.add( (*it).first);

				return return_list;
			}
			
			size_t size() const
			{
				return links.size();
			}

			bool HasFaultListSaturated() const
			{
				return links.size() > 64;
			}

			void AddTargets( const FaultCollection& fault_list)
			{
				for( _ag_imp_typename FaultCollection::const_iterator it = fault_list.begin();
					    it != fault_list.end(); it++)
				   AddFault(  (*it).first, (*it).second);
			
			}

			void AddFault( Link fault , size_t detected_time)
			{
				 assert( !HasFaultListSaturated() );
                 links[fault] = Block( detected_time);
			}

			void PrepareSequence( vector<bool>& omitted)
			{
				for( _ag_imp_typename LinkList::iterator it  = links.begin();
					     it != links.end(); it++)
				{
					(*it).second.PrepareSequence( omitted);
				}
			}

			void Search(const FaultCollection& detected_set,const vector<bool>& removed)
			{
#ifndef NDEBUG
				for( _ag_imp_typename  FaultCollection::const_iterator it = detected_set.begin();
					   it != detected_set.end(); it++)
				{
					assert(  links.find( (*it).first ) != links.end());
				}
#endif

				for( _ag_imp_typename  LinkList::iterator it = links.begin();
					   it != links.end(); it++)
				{
                     bool detected = detected_set.find( (*it).first ) != detected_set.end();
                     (*it).second.Search( detected, removed);                                          
				}

			}

			void CommitAndRemove( vector<bool> & omitted)
			{

                vector< Link > removables;
				for( _ag_imp_typename LinkList::iterator it = links.begin();
					  it != links.end(); it++)
				{
					if( (*it).second.IsCommitable())
					{
                        (*it).second.Commit(omitted);
						removables.push_back( (*it).first);
					}
				}

				for( _ag_imp_typename  vector<Link>::const_iterator it = removables.begin();
					   it != removables.end(); it++)
				{
					links.erase(*it);
				}
			
			}


		};


		template< class _TSimulator , class _Allocator>
		class BasicCompactor : public ProofsFaultSimulator< _TSimulator , _Allocator>   {

			typedef _Allocator Allocator;
			typedef _TSimulator TSimulator;
			typedef _ag_typename Allocator::T T;
			typedef _ag_typename Allocator::TCircuit TCircuit;
			typedef _ag_typename Allocator::WideInt WideInt;
			typedef _ag_typename Allocator::LogicValueManipulator  LogicValueManipulator;
			typedef _ag_typename Allocator::FeaturedGate  FeaturedGate;
			typedef _ag_typename Allocator::NodeList NodeList;
			typedef _ag_typename Allocator::ForwardLevelList ForwardLevelList;
			typedef _ag_typename Allocator::ReconvergentInputList   ReconvergentInputList ;
			typedef _ag_typename Allocator::LevelIndexKeyPair LevelIndexKeyPair;
			typedef _ag_typename Allocator::FaultCollection FaultCollection;
			typedef _ag_typename Allocator::FlipflopList FlipflopList;
		    typedef _ag_imp_typename _Allocator::FlipflopStoragePair   FlipflopStoragePair;
	        typedef _ag_imp_typename _Allocator::FlipflopStorage FlipflopStorage;
			typedef ProofsFaultSimulator< _TSimulator , Allocator>   TBaseEngine;
			typedef std::multimap<  double , string , less<double> > SortedPatternList;



		public:

			inline BasicCompactor(TSimulator & _simulator) : TBaseEngine(_simulator) {
				 parallel_faults= 64;

			}

			inline size_t get_VectorNumber() const
			{
				return vector_no;
			}

			
			inline size_t get_VectorsCount() const
			{
				return num_vectors;
			}

			inline void set_VectorsCount( size_t _num_vectors)
			{
                num_vectors = _num_vectors;
			}
		    
			bool after_eachVector(  const FaultCollection & fault_list,
				const FaultCollection & detected_set,
				string input_vector,
				size_t width) 
			{
				bool synchronized = true;
				if (find_sync_prefix && !sync_prefix_found)
				{
				  for(_ag_imp_typename IndicesList::const_iterator it = get_CircuitFlipflops().begin();
					      it != get_CircuitFlipflops().end(); it++)
				   {
					   if(  _IsLogicX(_virtual(this)->get_GoodMachineValueFor(*it)) )
					   {
						    synchronized = false;
						     break;
					   }
				  
				    }
				   if( synchronized)
				   {
					   sync_prefix_end = get_VectorNumber();
					   sync_prefix_found = true;
				   }
				}
				return TBaseEngine::after_eachVector( fault_list, detected_set, input_vector, width);
			}

			void BackupStates()
			{

				flipflop_states = TBaseEngine::get_DffList();
				TBaseEngine::CleanFlipflopStates();

			}
			void RestoreStates()
			{
				TBaseEngine::set_DffList(flipflop_states);
				TBaseEngine::CleanFlipflopStates();
			
			}

			void  RestoreAndOmitExceptSyncPrefix(const vector<bool>& initial_removed)
			{
				sync_prefix_end = 0;
				if( find_sync_prefix  && sync_prefix_found)
				{
				  total_restored = sync_prefix_end + 1; 
				  assert( sync_prefix_end < removed.size());
				  fill( removed.begin() , removed.begin() + sync_prefix_end, false); 
				}
				fill( removed.begin() + sync_prefix_end , removed.end(), true);

				for( vector<bool>::const_iterator it = initial_removed.begin();
					  it!= initial_removed.end(); it++)
					  if( !(*it))
						    *(removed.begin() + size_t(it - initial_removed.begin())) = false; 

				find_sync_prefix = false;

				cout << "sync prefix found at " << sync_prefix_end << "\n";
		        cout.flush();	
			}

			void WriteStepFile( ostream& step_file, bool step_file_write,
				                VectorCollection& vectors,
								FaultCollection& fault_list,
								FaultCollection& detected_set)
			{
				  size_t vec_size = 0;
				  if(vectors.size() != 0)
					  vec_size = vectors[0].size();
				
				  if(step_file_write)
				  {
  				     step_file  << detected_set.size()  << " " << fault_list.size() << " " << vectors.size() << " " << vec_size << std::endl;
					 vectors.Write(step_file);
					 //temp_fault_list.Sub(detected_faults, undetected_faults);
					 detected_set.Write( step_file , &Link::ToStringFaultInformation);
	                 fault_list.Write( step_file, &Link::ToStringFaultInformation);
				  }

			}



			
			void InitializeCompactor(ostream& step_file, bool step_file_write, FaultCollection& fault_list, VectorCollection& vectors, size_t width)
			{

				removed = vector<bool>( vectors.size());
				fill(removed.begin(), removed.end(), false);
				set_VectorsCount(vectors.size());
				find_sync_prefix = true;
				detected_list = vector<FaultCollection>( vectors.size() );
				sync_prefix_end = vectors.size() - 1;
				sync_prefix_found = false;
				original_set.add(fault_list);
				size_t fault_list_size = fault_list.size();
				detected_set = FaultCollection();
				original_detected = flipflopsaved_Run(fault_list, vectors, detected_list , detected_set, removed, width);
				WriteStepFile( step_file, step_file_write, vectors, fault_list, detected_set);

  				cout << "# initial vectors: " << vectors.size() 
					 << ", # faults detected: " << original_detected   
					 << "-rst_compaction\n"; 

			}


			void CreateUndetectedFaultList( vector<FaultCollection>& original_list, FaultCollection detected_set, 
				                            vector<FaultCollection>& undetected_list)
			{
				for( _ag_imp_typename vector<FaultCollection>::iterator it = original_list.begin();
					     it != original_list.end(); it++)
				{
					     FaultCollection undetected_;
					     (*it).Sub( detected_set, undetected_);
						 undetected_list.push_back( undetected_);
				}
			}

			

			void Print( size_t detected, size_t f_targ, const vector<bool>& removed)
			{
				size_t restored = 0;
				for(vector<bool>::const_iterator it =  removed.begin();  it != removed.end(); it++)
					restored += 1- size_t(*it) ;
                cout << "#  vectors: " << restored 
					<< ", #fault length " << f_targ << " # faults detected: " << detected    
					 << "-rst_compaction\n";
				cout.flush();


			}


			size_t RunFaultSimulator(FaultCollection& temp_set, vector<FaultCollection>& temp_detected_list, 
				                     FaultCollection& temp_detected_set,
									 const VectorCollection& vectors,
									 const vector<bool>& removed,
									 size_t width)
			{
			
				temp_detected_list = vector<FaultCollection>( vectors.size() );
				temp_detected_set = FaultCollection();
				temp_set = detected_set;

				size_t detected =  flipflopsaved_Run( temp_set, vectors, temp_detected_list, temp_detected_set,  removed, width);


				return detected;
	
			}

			
				struct nelem {
					double x;
					int index;

					nelem()
					{
						x = 0;
						index= 0;
					}

					nelem(double _x, int _index)
					{
						x =_x;
						index  = _index;
					}

					bool operator<(const nelem& b) const 
					{

						return x < b.x;
					}
				};

			void CreateMap(int len, vector<size_t>& from, vector<size_t>& to)
			{
			//	std::shuffle(


				std::vector< nelem >  ran(len);
				for( int i = 0; i < len; i++)
					ran[i] = nelem( Randomizer::RandomUniform() , i);
				std::sort(ran.begin(), ran.end(), less<nelem>());

				from  = std::vector<size_t>(len);
				to = std::vector<size_t>(len);

				for( int i = 0; i < len; i++)
				{
					 int j = ran[i].index;
					 from[i] = j;
					 to[j] = i;
				}
/*
				for( int i = 0; i < len; i++)
				{
					std::cout << i << ","  << from[i] << " ";
				}
*/
			}


			void PickTargetFaults(FaultCollection & F_targ, vector<FaultCollection>& temp_undetected_list, size_t n_faults)
			{
				_ag_imp_typename vector<bool>::const_iterator rem_it = removed.begin();
				_ag_imp_typename vector<FaultCollection>::iterator it;

				size_t _u_i = 0, u_i;
				for(; _u_i < temp_undetected_list.size();   _u_i++)
					{
#ifdef __RANDOM_ORDER_RESTORE__
					    u_i = to_map[_u_i];
#else
						u_i = temp_undetected_list.size()  - _u_i - 1;
#endif
						it = temp_undetected_list.begin() + u_i;
						rem_it = removed.begin() + u_i;
						if( (*it).size() !=0)  {
							_ag_imp_typename FaultCollection::iterator ty_it ;
							for( ty_it = (*it).begin();   ty_it != (*it).end(); ty_it++)
							{
								if(F_targ.size() == n_faults)
								  break;
	
								F_targ.add((*ty_it).first, u_i);




							//	cout << "considering  detection times :" << u_i << "\n";


                   
							}

							(*it).erase( (*it).begin(), ty_it);
							 
						}

						if( F_targ.size() == n_faults)
						 break;
						
					}

			}

			void AddTargets( FaultCollection& F_targ, FaultCollection & original_list, FaultCollection& new_faults)
			{
				for( _ag_imp_typename FaultCollection::iterator it = F_targ.begin();
					   it!= F_targ.end(); it++)
				{
					assert(  original_list.find((*it).first ) != original_list.end());
					size_t detected_time = original_list[ (*it).first ];
					(*it).second = detected_time;
				}

			    F_targ.add( new_faults);
			}


			void get_PossibleVectors(size_t u_i, vector<size_t>& detect_times )
			{
				while(true)
				{
					if(removed[u_i])
						detect_times.push_back(u_i);
					if( u_i == 0)
						break;
					u_i --;

				}

			}

			size_t pFunc( size_t u_i)
			{
				return u_i;

			}

			size_t RandomChooseDetectTime( vector<size_t>& detect_times)
			{
				 size_t sum = 0, old_sum;
				 for(size_t i  = 0; i < detect_times.size(); i++)
					  sum += pFunc( detect_times[i]);

				 size_t i_rand =  size_t(sum*Randomizer::RandomUniform());
				 old_sum = 0; sum = 0;
				  for(size_t i  = 0; i < detect_times.size(); i++)
				  {
					  sum += pFunc( detect_times[i]);
					  if( (i_rand >= old_sum) && (i_rand < sum))
						  return detect_times[i];
				  }

				  return 0;

			}


			void PrepareOmitSequence(FaultCollection& fault_list, FaultCollection& original_list, size_t max_time)
			{
				vector<size_t> detected_times;
				max_time = 0 ;
				_ag_imp_typename vector<size_t>::const_iterator end_detected_it;
				for( _ag_imp_typename FaultCollection::const_iterator it = fault_list.begin();
					   it!= fault_list.end(); it++)
				{
					assert(  original_list.find((*it).first ) != original_list.end());
					size_t detected_time = original_list[ (*it).first ];
					detected_times.push_back( detected_time  );	
					max_time = std::max( detected_time, max_time);
				}


				std::sort( detected_times.begin(), detected_times.end());
				end_detected_it = std::unique( detected_times.begin(), detected_times.end());

				for( _ag_imp_typename vector<size_t>::const_iterator  detect_it = detected_times.begin();
					    detect_it != end_detected_it ; detect_it++)
				{
/*#ifdef __RANDOM_ORDER_RESTORE__
					vector<size_t> possible_detect_times;
					get_PossibleVectors( *detect_it, possible_detect_times);
					size_t  u_i = RandomChooseDetectTime(  possible_detect_times);
					removed[ u_i ] = false;
#else
*/
				    size_t  u_i = (*detect_it);


					//cout  <<  u_i << " - ";
                    while(!removed[u_i])
					{
						if( u_i == 0) 
							 break;
						u_i --;
					}

				//	cout << u_i << ",";

					
					removed[ u_i] = false;
//#endif
				}

			}



			void Compact(FaultCollection& fault_list, 
				         VectorCollection& in_vectors,
				         VectorCollection& out_vectors, vector<bool>& initial_removed, 
						 ostream& step_file,
						 bool step_file_write,
						 size_t width)
			{
				 
				 FaultCollection temp_fault_list = fault_list;
				 analyzer_on = false;
				 analyzer_first_time = true;

				 bool no_future_compaction = false;
				 VectorCollection prev_vectors = in_vectors;
				 int n_iter = 0;
				 for(;;) {

				  VectorCollection temp_vectors;
				  FaultCollection temp_list = temp_fault_list;
				  FaultCollection detected_faults;
				  FaultCollection undetected_faults;
				  _Compact( temp_list, detected_faults, prev_vectors, temp_vectors, initial_removed,  step_file, step_file_write, width);



				  cout << "compacted ..." << n_iter++ << "\n";
				  cout.flush();

                  				  
				  
				  //break;

				  
				  if( prev_vectors.size() == temp_vectors.size())
					   break;
				  prev_vectors = temp_vectors;

				 }
				 /*
				 FaultCollection detected_faults;
				 FaultCollection undetected_faults;
				 VectorCollection temp_vectors;

                 CompactForHolding( temp_fault_list, detected_faults, prev_vectors, temp_vectors, initial_removed, width);
				 cout << "compacted ..." << n_iter++ << "\n";
				 cout.flush();
				 out_vectors = temp_vectors;
				 */
				 out_vectors = prev_vectors; 

				 
			}

			void set_ParallelFaults(size_t _parallel_faults)
			{
				parallel_faults = _parallel_faults;
			}
  
			size_t _Compact(FaultCollection& fault_list, FaultCollection & detected_fault_set, VectorCollection& in_vectors, 
				         VectorCollection& out_vectors, vector<bool>& initial_removed,
						  ostream& step_file,
						 bool step_file_write,
						 
						 size_t width)
			{
				FaultCollection temp_set;
				vector<FaultCollection> temp_detected_list;
				FaultCollection temp_detected_set;
				bool reset_detection = false;

				InitializeCompactor(step_file, step_file_write, fault_list, in_vectors, width);

				RestoreAndOmitExceptSyncPrefix(initial_removed);

				size_t ndetected = RunFaultSimulator( temp_set,  temp_detected_list, 
					                                  temp_detected_set, in_vectors, removed, width);

				detected_fault_set = temp_detected_set;

		
				for(;;)
				{
						if( ndetected == detected_set.size())
						break;


					vector<FaultCollection> temp_undetected_list;

					CreateUndetectedFaultList( detected_list, temp_detected_set, 
						                       temp_undetected_list);

					CreateMap( temp_undetected_list.size(), from_map, to_map);

					size_t  max_time = removed.size() - 1;

                    for(;;)
					{
						FaultCollection F_targ, F_targ_original ;
						//PickTargetFaults(F_targ, temp_undetected_list,64);
						//F_targ_original = F_targ;

						BlockedRestore< FaultCollection > Restorer;
						vector<bool>  _removed = removed;

						do
						{	
							FaultCollection detected_set;
							vector<FaultCollection> detected_list( in_vectors.size());
							FaultCollection F_targ_p;
							//size_t target_faults = rand() % 128;
							//if( target_faults < F_targ.size())
							//	  target_faults = F_targ.size();
							PickTargetFaults(F_targ_p, temp_undetected_list,parallel_faults - F_targ.size()); //- Restorer.size());
							AddTargets(F_targ, F_targ_original, F_targ_p);
							//Restorer.AddTargets( F_targ_p); 
							//F_targ = Restorer.get_FaultList();
							F_targ_original = F_targ;

							if( F_targ_original.size() == 0)
							{
								//removed  = _removed;
								break;
							}

							//AddTargets(F_targ, F_targ_original, F_targ_p);
							//F_targ_original = F_targ;

						    size_t detected = 
								//RunFaultSimulator( F_targ,  temp_detected_list, 
					            //                      temp_detected_set, in_vectors, removed, width);
							flipflopsaved_Run(F_targ, in_vectors, detected_list, detected_set, removed, width, max_time + 1);
                            //Restorer.Search(detected_set, removed);
							// Restorer.CommitAndRemove(removed);
							// _removed = removed;
						    //  Restorer.PrepareSequence(_removed);
							////removed = _removed;
		
							PrepareOmitSequence(F_targ, F_targ_original, max_time);

							//Print(detected,  Restorer.size(), removed);						
						
						} while( F_targ_original.size() != 0);

						if( F_targ_original.size() == 0)
							break;
					}

					cout << "Restored iteration end: testing. ...";


					ndetected = RunFaultSimulator(  temp_set, temp_detected_list, 
						                            temp_detected_set, in_vectors, removed, width);
					detected_fault_set = temp_detected_set;
					Print(ndetected, 0, removed);

#ifdef __CREATE_FLAKY_COMPACTOR__
					break;
#endif


					reset_detection = true;
			  
			}



				

				MakeVectorCollection( in_vectors, out_vectors, initial_removed);
				return ndetected;

			
			}

			void MakeVectorCollection(VectorCollection &inputs, VectorCollection& outputs,vector<bool>& initial_removed)
			{
				_ag_imp_typename vector<bool>::const_iterator rem_it = removed.begin();
				_ag_imp_typename vector<bool>::const_iterator in_rem_it = initial_removed.begin();
				vector<bool> _removed;
				for(_ag_imp_typename VectorCollection::const_iterator it= inputs.begin();
					    it != inputs.end(); it++)
				{
					if ( !(*rem_it)) {
						 outputs.push_back(*it);
						 _removed.push_back(*in_rem_it);
					}
				     rem_it++;
					 in_rem_it++;
				}

				assert(rem_it == removed.end());
				assert(in_rem_it == initial_removed.end());
				initial_removed = _removed;

			}

			//void Compact(FaultCollection& fault_list, VectorCollection& vectors, int width)
			//{

			//	InitializeCompactor( fault_list, vectors, width);
			//	RestoreAndOmitExceptSyncPrefix();


			//	  

			//	//printf("DFFs' are initialized at vector#: %d-rst_compaction\n",initial_stt_vec); 

   //            

		 //      backedup_Run( FaultCollection(detected_list), detected_set	

			//   CreateUndetectedFaultList( detected_list ,

			//	vec = numvec;
			//	vector<FaultCollection>::const_reverse_iterator j_it, k_it, next_it;
			//	j = detected_list.size();
			//
			//	for(j_it = detected_list.rbegin() ;  j_it!= detected_list.rend(); j_it--, j--)
			//	{
			//		if((*j_it).size() !=  0)
			//		{

			//			k_it = j_it; k = j;
			//			for(k_it--, k--; k != detected_list.rend() && (*k_it).size() == 0;k_it--, k--) ; //search for next flist
			//			next_it = k_it;
			//			next = k;
			//			while( 
			//				(tested_new = backedup_Run( FaultCollection(*j_it) ,  inputs , removed, width, total_restored))
			//				!= (*j_it).size()) && (vec > 0)) //initial_stt_vec) )
			//			{
			//				/* printf( "fl[%d]--#flts in list=%d--#detected=%d--vec#=%d--#Restore=%d\n",
			//				j, detectedflist[j][0],tested_new, vec, total_restored);
			//				*/ 
			//				if(vec >   j) vec = j;
			//				for(k=0;(k<3)&&(vec>0);k++) //restore 3 when set==0;
			//				{		 
			//					if(removed[vec])
			//					{
			//						removed[vec] = false;
			//						total_restored ++;
			//					}
			//					vec --;
			//					if(vec<next) k=3;
			//				}	
			//			}//while
			//		}//if(detectedflist...
			//		// Please add some print statements here!!
			//	}//for(j=...


			//	//printf("restored_old=%d\n",total_restored);
			//	if(1)
			//	{
			//		mark_tested =
			//			flipflopsaved_Run( FaultCollection(detected_set) ,  inputs , removed, width, total_restored-1); 
			//		for(k=0;k< inputs.size();k++)
			//		{
			//			if(!removed[k])
			//			{
			//				removed[k] = true;
			//				backup_states();
			//				tested_new =
			//					Run( FaultCollection(detected_set) ,  inputs , removed, width, total_restored-1);
			//				restore_states();
			//				
			//				if( tested_new < mark_tested)
			//					removed[k] = false;
			//				else
			//				{
			//					total_restored --;
			//				}
			//			}
			//		}
			//		detectedflist[numvec]=lastdflist;
			//	}
			//	//printf("restored_new=%d\n",total_restored);
			//	tested_new = flipflopsaved_Run( FaultCollection(detected_set) ,  inputs , removed, width, total_restored); 
			//	cout << "# compacted vectors:" << total_restored << " , # faults detected: "
			//		 << tested_net << "-rst_compaction\n"; 

			//	
			//	compact = 0;

			//	return(numvec-total_restored);
			//}

			inline size_t Run(FaultCollection & fault_list,
				              const VectorCollection& inputs,
				              const vector<bool>& removed,
							  size_t width,
							  size_t nthvector)
			{
				if( !removed[nthvector] )
				  return Run(fault_list, inputs[nthvector], width); 
			}

			inline size_t flipflopsaved_Run(FaultCollection & fault_list,
							  const VectorCollection& inputs,
							  vector<FaultCollection>& detected_list,				          
							  FaultCollection& detected_set,
							  const vector<bool>& removed,
							  size_t width,
							  size_t nvectors = size_t(-1))   
			{
		        BackupStates();
				TBaseEngine::ClearFaultSimulator();
				TBaseEngine::ClearNextIterationQueue();
				size_t detected = Run( fault_list, inputs, detected_list, detected_set,  removed, width, nvectors);
				TBaseEngine::ClearQueue();
				RestoreStates();
				return detected;

			}

			inline size_t Run(FaultCollection & fault_list, 
				              const VectorCollection& inputs,
							  vector<FaultCollection>& detected_list,
							  FaultCollection& detected_set,
							  const vector<bool>& removed,
							  size_t width,
							  size_t nvectors = size_t(-1))   
			{
				size_t total_faults = fault_list.size() ;
				size_t input_frame = 0;

				_ag_imp_typename vector<bool>::const_iterator rem_it = removed.begin();
				_ag_imp_typename vector<FaultCollection>::iterator detected_it = detected_list.begin();
				if( nvectors == size_t(-1))
					  nvectors = inputs.size();

				vector_no = 0;

				if(! _virtual(this)->before_Vectors(fault_list, inputs, width))
					return 0;

				for(VectorCollection::const_iterator it = inputs.begin();
					it != inputs.end(); it++, rem_it++,detected_it++)
				{
					assert(  rem_it != removed.end());
					if( nvectors-- == 0)
						 break;

					string _vector = *it;

					if( !(*rem_it) )
					{
						TBaseEngine::Run(fault_list, *detected_it , 
					  _vector, width);

	

				    detected_set.add(*detected_it);

					_virtual(this)->UpdateFlipflops();
					}

					vector_no ++ ;


					if(fault_list.empty())
						break;
				}

				if(! _virtual(this)->after_Vectors(fault_list, inputs, width))
					return 0;
				else
					return total_faults - fault_list.size();
			}

		
			inline size_t flipflopsaved_Run(FaultCollection & fault_list, 
				              const VectorCollection& inputs,
							  FaultCollection& detected_set,
							  const vector<bool>& removed,
							  size_t width,
							  size_t nvectors = size_t(-1))   	
			{

				BackupStates();
				TBaseEngine::ClearNextIterationQueue();
				TBaseEngine::ClearFaultSimulator();
				size_t detected=  Run( fault_list, inputs, detected_set, removed, width, nvectors);
				TBaseEngine::ClearQueue();
				RestoreStates();
				return detected;
			}

			void CompactorAnalyzer( size_t vector_no,
			                        FaultCollection& fault_list, 
				                    FaultCollection& fault_detected,
									string vector,
									size_t width)
			{
				if( analyzer_on) 
				{
					if(!analyzer_first_time )
					{
					  //if( removed[ vector_no ] )
						if( prev_vector == vector)
						{
						////	cout << "holding at " << vector_no << "\n";
							if( prev_fault_list_size == fault_list.size())
							{
								//cout << "\t no new faults detected \n";
								 if(  TBaseEngine::compare_FlipflopList( 
									     prev_flipflop_states, TBaseEngine::get_DffList()))
								{
								//	cout << "\t\t flip flop states no change \n";
									if( TBaseEngine::compare_FlipflopStates(
										   flipflop_for_next_iteration, prev_fault_propagation_list))
									{
										//cout << "\t\t\t propagation list match \n";
										removed[ vector_no ] = true;
									}

								}
					 		}
						}
					}
				prev_fault_list_size = fault_list.size();
				prev_vector = vector;
				prev_flipflop_states = TBaseEngine::get_DffList();
				TBaseEngine::DeleteFlipflopStates( prev_fault_propagation_list);
				prev_fault_propagation_list.clear();
				TBaseEngine::copy_FlipflopStates( flipflop_for_next_iteration, prev_fault_propagation_list);
				analyzer_first_time = false;
				}

			}

			void MergeRemovable( vector<bool>& initial_removed)
			{
				assert( initial_removed.size() == removed.size());
				for( size_t i = 0; i < initial_removed.size(); i++)
					 if( !initial_removed[i])
						   removed[i] = false;
                
			}

			void CompactForHolding( FaultCollection& fault_list, 
									FaultCollection & detected_fault_set, 
									VectorCollection& in_vectors, 
				                    VectorCollection& out_vectors, 
									vector<bool>& initial_removed,  
									size_t width)
			{
				analyzer_on = true;
				analyzer_first_time = true;
				FaultCollection detected_faults;
				removed = vector<bool>( initial_removed.size(), false);
				size_t ndetected = flipflopsaved_Run(fault_list, in_vectors, detected_faults, removed, width);
				MergeRemovable(initial_removed);
				Print(ndetected, 0, removed);
				MakeVectorCollection( in_vectors, out_vectors, initial_removed);
				analyzer_on =  false;


			}
				                    




			inline size_t Run(FaultCollection & fault_list, 
				              const VectorCollection& inputs,
							  FaultCollection& detected_set,
							  const vector<bool>& removed,
							  size_t width,
							  size_t nvectors = size_t(-1))   
			{
				size_t total_faults = fault_list.size() ;
				size_t input_frame = 0;


				vector_no = 0;

				_ag_imp_typename vector<bool>::const_iterator rem_it = removed.begin();
				if( nvectors == size_t(-1))
					  nvectors = inputs.size();

				if(! _virtual(this)->before_Vectors(fault_list, inputs, width))
					return 0;

				for(VectorCollection::const_iterator it = inputs.begin();
					it != inputs.end(); it++, rem_it++)
				{
					assert(  rem_it != removed.end());
					if( nvectors-- == 0)
						 break;

					FaultCollection faults_detected;

					string _vector = (*it);

					if( !(*rem_it) )
					{
						TBaseEngine::Run(fault_list, faults_detected ,_vector, width);
						_virtual(this)->CompactorAnalyzer(vector_no, fault_list, faults_detected, _vector, width);
						_virtual(this)->UpdateFlipflops();


				    detected_set.add(faults_detected);
					}

					vector_no++;




					if(fault_list.empty())
						break;
				}

				if(! _virtual(this)->after_Vectors(fault_list, inputs, width))
					return 0;
				else
					return total_faults - fault_list.size();
			}

			



		protected:
			 vector<bool> removed;
			 bool find_sync_prefix, sync_prefix_found;
			 FlipflopList flipflop_states, prev_flipflop_states;
			 FlipflopStorage prev_fault_propagation_list;
			 size_t vector_no, sync_prefix_end, total_restored, num_vectors;
			 size_t original_detected;
			 size_t prev_fault_list_size;
			 string prev_vector;
			 bool analyzer_on, analyzer_first_time;
			 size_t parallel_faults;
			 FaultCollection original_set,detected_set;
			 vector<FaultCollection> detected_list;
			 vector<size_t> from_map, to_map;
			


	

		};

	}

  }



#endif

