
#include <stdio.h>
//#include <stdlib.h>
#include "parameter.h"
#include "define.h"
#include "macro.h"
#include "assert.h"

#define HEADER 0xc1c5c6c5
#define INVALID_HEADER 0xcdcdcd1c
int _MEM = 0;
static int mallocs = 0,frees = 0;
int saves;
extern int _FSIM;
void print_mem_state(char *a)
{
#ifdef HOPE_MEM_LEAK_DEBUG
	printf("%s MEM = %d m= %d f =%d a = %d fsim = %d\n",a, _MEM, mallocs, frees, _MEM/( mallocs - frees), _FSIM );
	fflush(stdout);
#endif
}


typedef struct MEMBLOCK * MEMBLOCKPTR;
typedef struct MEMBLOCK {
	int size;
    int header;
	char *file;
	int line;
	int retag;
	int fsim;
	MEMBLOCKPTR next;
	MEMBLOCKPTR prev;
} MEMBLOCK;

static MEMBLOCK *mem_head = NULL, *mem_tail = NULL;


int add_block( char *p, char * file, int line, int size)
{
	MEMBLOCK *mem = (MEMBLOCK *)p; 
	mem->file = file;
	mem->size = size;
	mem->line = line;
	mem->header = HEADER;
	mem->next = NULL;
	mem->retag = 0;
	mem->fsim = _FSIM;
	if( mem_tail != NULL)
	{
  	  mem->prev = mem_tail;
	  mem_tail->next = mem;
	  mem_tail = mem;
	} else
	{
	  mem_tail = mem;
	  mem_head = mem;
	  mem_head->prev = NULL; 
	}
	return size;
}

int remove_block( char *p )
{
	 MEMBLOCK *mem = (MEMBLOCK *) p;
	 int size;
	 assert( mem->header == HEADER);
	 assert( mem->size >= 0);
	 size = mem->size;

	 if(  (mem->prev != NULL) && (mem->next!= NULL))
	 {
		 mem->prev->next = mem->next;
		 mem->next->prev = mem->prev;
	 }
	 else if( (mem->prev == NULL) && (mem->next == NULL))
	 {
		 mem_head = NULL;
		 mem_tail = NULL;
	 }
	 else if( (mem->prev != NULL) && (mem->next == NULL))
	 {
		 mem->prev->next = NULL;
		 mem_tail = mem->prev;
	 }
	 else if( (mem->prev == NULL) && (mem->next != NULL))
	 {
		 mem->next->prev = NULL;
		 mem_head = mem->next;
	 }

	 mem->size = -1;
	 mem->header = INVALID_HEADER;
	 return size;
}

volatile char *hope_malloc( char *file, int line, int size)
{
	int _size = size + sizeof(MEMBLOCK);
	char *mem;
	_MEM += _size ;
    mem = (char *) malloc(_size);
    add_block( mem  , file, line , _size);

	mallocs++;
	return (char *) ( mem + sizeof(MEMBLOCK));
}

volatile char *hope_realloc( char *file, int line, char *p, int size)
{
	
	int _size = size + sizeof(MEMBLOCK), old;
	char *mem;
	assert(0);
	old = remove_block( p - sizeof(MEMBLOCK));

	_MEM +=  -old + _size ;

	mem = (char *) realloc( (p - sizeof(MEMBLOCK)),_size);
	add_block( mem , file, line, _size);

	mallocs++; frees++;
	//*( (MEMBLOCK *) mem).size = _size;

	return (char *) (mem + sizeof(MEMBLOCK));
}


void print_memblocks(char *memfile)
{
#ifdef HOPE_MEM_LEAK_DEBUG
	MEMBLOCK *temp = mem_head;
   FILE *fp = fopen(memfile,"w");
   int t = 1;
   while( temp != NULL)
   {
	   fprintf(fp,"%d. %s %d %d %d %d %d\n", t, temp->file, temp->fsim , _FSIM, temp->line, temp->size,temp->retag);
	   t++;
	   temp = temp->next;

   }
   fclose(fp);
#endif
}

void retag_memblock(char * file, int line, char *p)
{
	MEMBLOCK *mem = (MEMBLOCK *)(  p  - sizeof(MEMBLOCK));
	assert( mem->header == HEADER);
	mem->file = file;
	mem->line = line;
	mem->retag++;
	mem->fsim = _FSIM;
}

volatile void hope_free( char *file, int line, char *p)
{
	//assert(0);
	int size = 	remove_block( p  - sizeof(MEMBLOCK));
	_MEM -= size;
	frees++;
	free( (p - sizeof(MEMBLOCK)) );
}

volatile char * hope_calloc( char *file, int line, int s, int v)
{
     int _size = s*v + sizeof(MEMBLOCK);
   char *mem;
   assert(0);

   _MEM += _size ;
   mem = (char *) calloc(_size , 1);
   add_block( mem, file, line, _size);
   mallocs++;
   return (char *) (mem + sizeof(MEMBLOCK)); 

}


//
//volatile char *hope_malloc( char *file, int line, int size)
//{
//	int _size = size + 2*sizeof(int);
//	char *mem;
//	_MEM += _size - sizeof(int);
//	mem = (char *) malloc(_size);
//	/**( (MEMBLOCK*) mem ).size = _size;
//	*( (MEMBLOCK*) mem ).file = file;
//	*( (MEMBLOCK*) mem ).line = line;
//	*/
//	*( (int *) mem) = _size; // - sizeof(int);
//	*( (int *) (mem + sizeof(int))) = HEADER;
//	mallocs++;
//	return (char *) ( mem + 2*sizeof(int));
//}
//
//volatile char *hope_realloc( char *file, int line, char *p, int size)
//{
//	//assert(0);
//	int _size = size + 2*sizeof(int);
//	char *mem;
//	_MEM +=  -*((int *) (p - 2*sizeof(int))) + _size - sizeof(int);
//	assert( *((int *) (p - sizeof(int)) ) == HEADER);
//	mem = (char *) realloc( (p - 2*sizeof(int)),_size);
//	*( (int *) mem) = _size;
//    *( (int *)( mem + sizeof(int))) = HEADER;
//	mallocs++; frees++;
//	//*( (MEMBLOCK *) mem).size = _size;
//
//	return (char *) (mem + 2*sizeof(int));
//}
//
//volatile void hope_free( char *file, int line, char *p)
//{
//	//assert(0);
//	 _MEM -=  *((int *) (p - 2*sizeof(int)) ) - sizeof(int);
//	 assert( *((int *) (p - 2*sizeof(int)) ) >= 0);
//	 assert( *((int *) (p - sizeof(int)) ) == HEADER);
//     frees++;
//     *((int *) (p - 2*sizeof(int)) ) = -1;
//	 *((int *) (p - sizeof(int)) ) = rand() ;
//	 free( (p - 2*sizeof(int)) );
//}
//
//volatile char * hope_calloc( char *file, int line, int s, int v)
//{
//   int _size = s*v + 2*sizeof(int);
//   char *mem;
//   _MEM += _size - sizeof(int);
//   mem = (char *) calloc(_size , 1);
//   *( (int *) mem) = _size;
//   *( (int *) (mem + sizeof(int))) = HEADER;
//   mallocs++;
//   return (char *) (mem + 2*sizeof(int)); 
//
//}
